module GeneFinder

using BioSequences:
    DNA,
    NucleicAcidAlphabet,
    DNAAlphabet,
    AminoAcidAlphabet,
    LongDNA,
    LongAA,
    LongSequence,
    LongSubSeq,
    @biore_str,
    GeneticCode,
    reverse_complement
using FASTX: FASTA, sequence, FASTAReader
using IterTools: takewhile, iterated
using MarkovChainHammer.Trajectory: generate
using PrecompileTools
using StatsBase: countmap
using TestItems: @testitem

include("types.jl")
export ORF, CDS, Protein, TCM, TPM, TransitionModel

include("algorithms/findorfs.jl")
export locationiterator, findorfs, getorfdna, getorfaa

include("io.jl")
export write_cds, write_proteins, write_bed, write_gff

include("biomarkovchains.jl")
export transition_count_matrix,
    transition_probability_matrix, 
    initial_distribution, 
    transition_model,
    sequenceprobability,
    iscoding,
    perronfrobenius,
    generatednaseq

include("helpers.jl")
export fasta_to_dna,
    nucleotidefreqs,
    dinucleotides,
    hasprematurestop

include("models/models.jl")
export ECOLICDS, ECOLINOCDS

include("extended.jl")

@setup_workload begin
    # Putting some things in `@setup_workload` instead of `@compile_workload` can reduce the size of the
    # precompile file and potentially make loading faster.
    using BioSequences
    seq = randdnaseq(10^3)
    @compile_workload begin
        # all calls in this block will be precompiled, regardless of whether
        # they belong to your package or not (on Julia 1.8 and higher)
        findorfs(seq)
        getorfdna(seq)
        getorfaa(seq)
        nucleotidefreqs(seq)
        dinucleotides(seq)
        sequenceprobability(seq, ECOLICDS)
        perronfrobenius(seq)
    end
end

end
